'''
Created on: Mar 04, 2014

@author: qwang

This module provide api to use youpai cloud

API provided:
    1. get_access_url(url, size)
    2. put(url, data = None)
    3. delete(url)
    4. get_count()
    5. get_usage()
    6. get_info(url)
'''
import upyun

import configure

from tool import download, gen_key, encode
from exception import ParamException, ConfigParamInvalidException, \
        UpYunServiceException, UpYunClientException

class UpyunClient(object):
    '''
    Implement youpai service api
    '''
    def __init__(self, config):
        '''
        initialize UpyunApi object

        @Param: 
            1. config: upyun client configuration which is supposed to \
                    contain BUCKET, USERNAME, PASSWORD, TIMEOUT. \
        '''
        self._check(config)
        self._config = config
        self._upyun_client = upyun.UpYun(config['BUCKET'], config['USERNAME'], \
                config['PASSWORD'], timeout=config['TIMEOUT'], \
                endpoint=upyun.ED_AUTO)
        self._server = None
        if self._config.has_key('SERVER'):
            self._server = self._config['SERVER']

    def put(self, url, data = None):
        '''
        upload image to youpai cloud

        @Param: 
            1. url to be upload
            2. data of image. Default to be None, so API support downloading \
                    data by url, and uploading to cloud automatically. \
                    Also, you can provide url as well as data

        @output:
            1. when successed
                a dictionary like:
                    dict = {
                        'width': 50,
                        'height': 50,
                        'frames': 1,
                        'file-type': 'JPEG' 
                        }
                as well as "access_url" for original image
            2. when failed
                exception raised
                refers to: 
                    http://wiki.upyun.com/index.php?title=HTTP_REST_API%E6%8E%A5%E5%8F%A3#.E6.A0.87.E5.87.86API.E9.94.99.E8.AF.AF.E4.BB.A3.E7.A0.81.E8.A1.A8    

        @Notice:
            There is no way to determin if an image(url) to be upload \
                    has been uploaded before.
            So, try not to upload duplicated image !!!
        '''
        url = url.strip()
        if not data:
            data = download(url)
        key = gen_key(url)
        res = None
        try:
            res = self._upyun_client.put(key, data, checksum=True)
        except upyun.UpYunServiceException, error:
            raise UpYunServiceException(error)
        except upyun.UpYunClientException, error:
            raise UpYunClientException(error)
        res['access_url'] = self.get_access_url(url)
        return res

    def delete(self, url):
        '''
        delete image
        '''
        key = gen_key(url)
        try:
            self._upyun_client.delete(key)
        except upyun.UpYunServiceException, error:
            raise UpYunServiceException(error)
        except upyun.UpYunClientException, error:
            raise UpYunServiceException(error)

    def get_access_url(self, url, size = None):
        '''
        get url, by which you can access image in youpai cloud

        @param:
            url: original image url
            size: supported sizes defined in configure.py. 
                Default None, means original image
        '''
        url_youpai = None

        if size and size not in configure.IMAGE_WIDTH:
            raise ParamException('unsupported size: %s' % size)

        key = encode(url)
        if not self._server:
            url_youpai = configure.SERVER % self._config['BUCKET'] + key
        else:
            url_youpai = 'http://' + self._server + '/' + key

        if size == 'original':
            size = None

        if size:
            url_youpai += ('!' + size)

        return url_youpai

    def get_count(self):
        '''
        get image count in youpai cloud
        '''
        try:
            infos = self._upyun_client.getlist()
        except upyun.UpYunServiceException, error:
            raise UpYunServiceException(error)
        except upyun.UpYunClientException, error:
            raise UpYunClientException(error)

        return len(infos)

    def get_info(self, url):
        '''
        get image info in youpai cloud
        '''
        key = gen_key(url)
        try:
            info = self._upyun_client.getinfo(key)
        except upyun.UpYunServiceException, error:
            raise UpYunServiceException(error)
        except upyun.UpYunClientException, error:
            raise UpYunClientException(error)
        return info

    def get_usage(self):
        '''
        get the space used, unit: B
        '''
        try:
            usage = self._upyun_client.usage()
        except upyun.UpYunServiceException, error:
            raise UpYunServiceException(error)
        except upyun.UpYunClientException, error:
            raise UpYunClientException(error)
        return usage

    def _check(self, config):
        '''
        check config
        '''

        if not config or not isinstance(config, dict):
            raise ConfigParamInvalidException('CONFIG_INVALID, dictionay needed')
        for item in configure.youpai_config:
            if item not in config:
                raise ConfigParamInvalidException('CONFIG_INVALID %s needed' % item)
        if 'TIMEOUT' not in config:
            config['TIMEOUT'] = configure.TIMEOUT



'''
Created on Mar 11, 2014

@author: zxzhang

'''

import base64

import downloader

from exception import DownloadException

def gen_key(url):
    '''
    This function is to generate key, used for upload image into cloud

    input: url of image
    output: key of image

    '''
    key = encode(url)
    return key

def download(url):
    '''
    This function is to download image.

    Note1:
        1. suppose url inputed refers to a real image, not a html or other object
        2. check image when downloaded, check item:
            a. an image ?
            b. image type supported ?
            c. downloaded successfully ?
    Note2: 
        check image was deprecated
    '''
    resp = _download(url)
    if not resp:
        raise DownloadException(resp, 'Download failed, try again later!')
    if resp.status == 200:
        return resp.body
    else:
        raise DownloadException(resp)

def encode(s):
    '''
    base64 encode
    '''
    if isinstance(s, unicode):
        s = s.encode('utf-8')
    url_encoded = base64.urlsafe_b64encode(s)
    return url_encoded

def _download(url):
    '''
    download image by url
    '''
    max_try_count = 3
    try_count = 0
    resp = None
    while True:
        if try_count >= max_try_count:
            break
        try:
            resp = downloader.download(url)
        except Exception:
            try_count += 1
            continue
        try_count += 1
    return resp



#! /usr/bin/env python
'''
Created on Mar 11, 2014

Author: zxzhang

This module defines specific exceptions.

'''

import configure

class DownloadException(Exception):

    def __init__(self, resp = None, msg = None):
        self.resp = resp
        if resp:
            self.args = (resp.status, resp.headers)
            self.status = resp.status
            self.body = resp.body
            self.headers = resp.headers
        self.msg = msg
    def __str__(self):
        if not self.resp:
            return self.msg
        else:
            return 'status: %s \n headers: %s' % \
                (self.status, self.headers)

class ParamException(Exception):

    def __init__(self, msg = ''):
        self.args = msg
        self.msg = msg
        self.msg += '\nsupported image size: '
        for size in configure.IMAGE_WIDTH:
            self.msg += ('%s: %d\n' % (size, configure.IMAGE_WIDTH[size]))

    def __str__(self):
        return self.msg

class ConfigParamInvalidException(Exception):

    def __init__(self, msg = ''):
        self.args = msg
        self.msg = msg
        self.msg += '''\nconfig is supposed to be dictionary type, \
                    including 'BUCKET', 'USERNAME', 'PASSWORD' keys at least'''
    
    def __str__(self):
        return self.msg


class UpYunServiceException(Exception):
    def __init__(self, error):
        self.args = (error.status, error.msg, error.err)
        self.status = error.status
        self.msg = error.msg
        self.err = error.err


class UpYunClientException(Exception):
    def __init__(self, error):
        self.args = (error.msg)
        self.msg = error.msg




'''
Created on Mar 11, 2014

Author: zxzhang

'''

# client conf
SERVER = 'http://%s.b0.upaiyun.com/'

# supported image types
valid_image_types = ['image/jpg', 'image/png', 'image/gif',\
        'image/bmp', 'image/tif', 'image/jpeg']

# supported sizes
IMAGE_WIDTH = {
    'big': 500,
    'mid': 240,
    'small': 120,
    'original': None
}

# time out
TIMEOUT = 60

youpai_config = ['BUCKET', 'USERNAME', 'PASSWORD']



'''
Created on Aug 14, 2012

@author: Yu
'''
import urllib3, urlparse, urllib, time
from time import time
import random

_SITES_RATE_LIMIT = {             
              }

_SITES_LAST_ACCESS = {}

_NUM_POOLS = 10
_TIMEOUT = 30
_DEFAULT_HEADER =  {
        'Accept' :"text/html,application/xhtml+xml,application/xml;q=0.9,*/*;q=0.8",
        'Accept-Charset' : "ISO-8859-1,utf-8;q=0.7,*;q=0.3",
        'Accept-Encoding' : "gzip,deflate,sdch",
        'Accept-Language' : "en-US,en;q=0.8",
        'Cache-Control' : 'max-age=0',
        'Connection' :'keep-alive',
        'User-Agent' :"Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/535.19 (KHTML, like Gecko) Chrome/18.0.1025.162 Safari/535.19"
    } 

_HTTP = urllib3.PoolManager(num_pools = _NUM_POOLS, timeout = _TIMEOUT)
_PROXY = {}

class Response(object):
    def __init__(self, status = 200, body = None, headers = None):
        self.status = status
        self.body = body
        self.headers = None or {}

def download(url, method = "GET", params = None, body = None, headers = None, proxys = None):
    params = params or {}
    headers = headers or {}
    proxys = proxys or []
    if isinstance(url, unicode):
        url = url.encode('utf8')
    url = urllib.quote(url, safe=";|/|?|:|@|&|=|+|$|,|%")
    for k, v in params.iteritems():
        if isinstance(v, unicode):
            params[k] = v.encode('utf8')    
    
    if params:
        if method == "GET" and url.find('?') == -1:            
            url = url + "?" + urllib.urlencode(params)
        elif method == "POST":
            body = urllib.urlencode(params)
    
    parse = urlparse.urlparse(url)        
    site = parse.netloc
        
    if _SITES_RATE_LIMIT.has_key(site):
        last_visit = _SITES_LAST_ACCESS.get(site)
        now = int(time())        
        if last_visit:
            waittime = last_visit + _SITES_RATE_LIMIT[site] - now
            if waittime > 0:
                time.sleep(waittime)
        _SITES_LAST_ACCESS[site] = now                
    
    
    proxy = proxys[random.randint(0, len(proxys) - 1)] if proxys else None

    pool = _HTTP
    if proxy:        
        if not _PROXY.has_key(proxy):
            _PROXY[proxy] = urllib3.proxy_from_url(proxy)
        pool = _PROXY[proxy]
    _headers = _DEFAULT_HEADER.copy()
    _headers["Host"] = parse.netloc
    for k, v in headers.iteritems():
        if not v and _headers.has_key(k):
            _headers.pop(k)
        else:
            _headers[k] = v
    try:
        resp = pool.urlopen(method = method, url = url, body = body, headers = _headers)
        resp = process_response(resp)
        if resp.status in [301, 302]:
            redirect_url = resp.headers.get('location')
            resp = download(url = redirect_url, method = method, params = params, body = body, headers = headers, proxys = proxys)            
    except Exception:
        raise
    return resp 
    
def process_response(resp):
    status = resp.status
    headers = resp.headers
    data = resp.data

    if status == 200:
        charset = None
        m = headers.get('content-type')            
        if m:
            fields = m.split(';')
            if len(fields) > 1:
                k, v = fields[1].split('=')
                if k.lower().strip() == 'charset':
                    charset = v.strip()

        if charset:
            try:
                data = unicode(data, charset)
            except:
                if charset.lower() == 'gb2312':
                    #try to decode as gbk
                    try:
                        data = unicode(data, 'gbk')
                    except:
                        pass
    
    return Response(status = resp.status, headers = headers, body = data) 



#!/usr/bin/env python
import os
try:
    from setuptools import setup, Extension
except ImportError:
    from distutils.core import setup, Extension

sdict = {
    'name' : 'py-upyun',
    'version' : '0.1.3',
    'description' : 'Youpai Usage API',
    'author' : 'zixuan zhang',
    'author_email' : 'zixuan.zhang.victor@gmail.com',
    'keywords' : ['youpai', 'upyun', 'cloud'],
    'packages' : ['youpai'],
    'install_requires': ['urllib3>=1.7'],
    'classifiers' : [
        'Environment :: Console',
        'Intended Audience :: Developers',
        'Operating System :: OS Independent',
        'Programming Language :: Python'],
}

setup(**sdict)




#! /usr/bin/env python
# -*- coding: utf-8 -*-

import os
import socket
import urllib
import hashlib
import datetime

HTTP_EXTEND = None

try:
    import requests
    HTTP_EXTEND = True
except ImportError:
    import httplib

__version__ = '2.2.1'

ED_LIST = ['v%d.api.upyun.com' % ed for ed in range(4)]
ED_AUTO, ED_TELECOM, ED_CNC, ED_CTT = ED_LIST


# wsgiref.handlers.format_date_time

def httpdate_rfc1123(dt):
    """Return a string representation of a date according to RFC 1123
    (HTTP/1.1).

    The supplied date must be in UTC.

    """
    weekday = ["Mon", "Tue", "Wed", "Thu", "Fri", "Sat", "Sun"][dt.weekday()]
    month = ["Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul", "Aug", "Sep",
             "Oct", "Nov", "Dec"][dt.month - 1]
    return "%s, %02d %s %04d %02d:%02d:%02d GMT" % \
        (weekday, dt.day, month, dt.year, dt.hour, dt.minute, dt.second)


class UpYunServiceException(Exception):
    def __init__(self, status, msg, err):
        self.args = (status, msg, err)
        self.status = status
        self.msg = msg
        self.err = err


class UpYunClientException(Exception):
    def __init__(self, msg):
        self.args = (msg)
        self.msg = msg


class UpYun:

    def __init__(self, bucket, username, password,
                 timeout=None, endpoint=None):
        self.bucket = bucket
        self.username = username
        self.password = hashlib.md5(password).hexdigest()
        self.timeout = timeout or 60
        self.endpoint = endpoint or ED_AUTO
        self.user_agent = None

        if HTTP_EXTEND:
            self.session = requests.Session()

    # --- public API

    def usage(self, key='/'):
        return self.__do_http_request('GET', key, args='?usage')

    def put(self, key, value, checksum=False, headers=None):
        """
        >>> with open('foo.png', 'rb') as f:
        >>>    res = up.put('/path/to/bar.png', f, checksum=False,
        >>>                 headers={"x-gmkerl-rotate": "180"}})
        """
        if headers is None:
            headers = {}
        headers['Mkdir'] = 'true'
        if checksum is True:
            headers['Content-MD5'] = self.__make_content_md5(value)
        h = self.__do_http_request('PUT', key, value, headers)

        return self.__get_meta_headers(h)

    def get(self, key, value=None):
        """
        >>> with open('bar.png', 'wb') as f:
        >>>    up.get('/path/to/bar.png', f)
        """
        return self.__do_http_request('GET', key, of=value)

    def delete(self, key):
        self.__do_http_request('DELETE', key)

    def mkdir(self, key):
        headers = {'Folder': 'true'}
        self.__do_http_request('POST', key, headers=headers)

    def getlist(self, key='/'):
        content = self.__do_http_request('GET', key)
        if content == "":
            return []
        items = content.split('\n')
        return [dict(zip(['name', 'type', 'size', 'time'],
                x.split('\t'))) for x in items]

    def getinfo(self, key):
        h = self.__do_http_request('HEAD', key)
        return self.__get_meta_headers(h)

    # --- private API

    def __do_http_request(self, method, key,
                          value=None, headers=None, of=None, args=''):

        uri = '/' + self.bucket + (lambda x: x[0] == '/' and x or '/'+x)(key)
        if isinstance(uri, unicode):
            uri = uri.encode('utf-8')

        uri = urllib.quote(uri, safe="~/") + args

        if headers is None:
            headers = {}

        length = 0
        if isinstance(value, file):
            length = os.fstat(value.fileno()).st_size
        elif isinstance(value, str):
            length = len(value)
            headers['Content-Length'] = length
        elif value is not None:
            raise UpYunClientException("object type error")

        # Date Format: RFC 1123
        dt = httpdate_rfc1123(datetime.datetime.utcnow())
        signature = self.__make_signature(method, uri, dt, length)

        headers['Date'] = dt
        headers['Authorization'] = signature
        if self.user_agent:
            headers['User-Agent'] = self.user_agent
        else:
            headers['User-Agent'] = self.__make_user_agent()

        if HTTP_EXTEND:
            return self.__do_http_extend(method, uri, value, headers, of)
        else:
            return self.__do_http_basic(method, uri, value, headers, of)

    def __make_signature(self, method, uri, date, length):
        signstr = '&'.join([method, uri, date, str(length), self.password])
        signature = hashlib.md5(signstr).hexdigest()
        return 'UpYun ' + self.username + ':' + signature

    def __make_user_agent(self):
        default = "upyun-python-sdk/" + __version__

        if HTTP_EXTEND:
            return "%s %s" % (default, requests.utils.default_user_agent())
        else:
            return default

    def __make_content_md5(self, value):
        if isinstance(value, file):
            md5 = hashlib.md5()
            for chunk in iter(lambda: value.read(8192), b''):
                md5.update(chunk)
            value.seek(0)
            return md5.hexdigest()
        elif isinstance(value, str):
            return hashlib.md5(value).hexdigest()
        else:
            raise UpYunClientException("object type error")

    def __get_meta_headers(self, headers):
        return dict(iter([(k[8:].lower(), v) for k, v in headers
                    if k[:8].lower() == 'x-upyun-']))

    # http://docs.python.org/2/library/httplib.html

    def __do_http_basic(self, method, uri,
                        value=None, headers=None, of=None):

        content, msg, err, status = None, None, None, None
        try:
            connection = httplib.HTTPConnection(self.endpoint,
                                                timeout=self.timeout)
            # connection.set_debuglevel(1)
            connection.request(method, uri, value, headers)
            response = connection.getresponse()

            status = response.status
            if status / 100 == 2:
                if method == "GET" and of:
                    while True:
                        chunk = response.read(8192)
                        if not chunk:
                            break
                        of.write(chunk)
                if method == "GET" and of is None:
                    content = response.read()
                if method == "PUT" or method == "HEAD":
                    content = response.getheaders()
            else:
                msg = response.reason
                err = response.read()

        except (httplib.HTTPException, socket.error, socket.timeout) as e:
            raise UpYunClientException(str(e))
        except Exception as e:
            raise UpYunClientException(str(e))
        finally:
            if connection:
                connection.close()

        if msg:
            raise UpYunServiceException(status, msg, err)

        return content

    # http://docs.python-requests.org/

    def __do_http_extend(self, method, uri,
                         value=None, headers=None, of=None):

        content, msg, err, status = None, None, None, None
        URL = "http://" + self.endpoint + uri
        requests.adapters.DEFAULT_RETRIES = 5

        try:
            response = self.session.request(method, URL, data=value,
                                            headers=headers,
                                            timeout=self.timeout)
            status = response.status_code
            if status / 100 == 2:
                if method == "GET" and of:
                    for chunk in response.iter_content(8192):
                        if not chunk:
                            break
                        of.write(chunk)
                elif method == "GET" and of is None:
                    content = response.content
                elif method == "PUT" or method == "HEAD":
                    content = response.headers.items()
            else:
                msg = response.reason
                err = response.content

        except requests.exceptions.ConnectionError as e:
            raise UpYunClientException(str(e))
        except requests.exceptions.RequestException as e:
            raise UpYunClientException(str(e))
        except Exception as e:
            raise UpYunClientException(str(e))

        if msg:
            raise UpYunServiceException(status, msg, err)

        return content

if __name__ == '__main__':
    pass
